from abc import ABC, abstractmethod


class SelectByHiddenValueMixin(ABC):
    """
    mixin class that describes that this element has selectable elements that can be selected-by-value
    """

    @abstractmethod
    def select_by_value(self, value: object) -> None:
        """
        This method selects the element by its value
        :param value: the value to select
        """



from abc import ABC, abstractmethod


class SelectByVisibleTextMixin(ABC):
    """
    mixin class that describes that this element has selectable elements that can be selected-by-visible-text
    """
    @abstractmethod
    def select_by_text(self, visible_text: str):
        """
        This method selects the element by visible text
        :param visible_text: the visible text of the element that should be selected
        """



from abc import ABC, abstractmethod
from typing import TypeVar
import time

ClickableMixinTypeT = TypeVar("ClickableMixinTypeT", bound="ClickableMixin")


class ClickableMixin(ABC):
    """
    mixin class that describes a clickable element
    """

    @abstractmethod
    def click(self) -> None:
        """
        Executes a click at the element
        """

    @abstractmethod
    def is_clickable(self) -> bool:
        """
        This method checks if the element is theoretically clickable. This means, that no other element does cover it
        up.

        :return: True if the element should be theoretically clickable, False otherwise.
        """

    def is_clickable_within(self, time_sec: float) -> bool:
        """
        This method waits for a maximum of `time_sec` seconds for the element to be clickable. If the element is not
        clickable within the provided time `time_sec`, the method returns False.

        :param time_sec: the maximum time to wait for this element to be clickable
        :return: True if it is clickable, otherwise False
        """
        start_time = time.perf_counter()
        while time.perf_counter() - start_time < time_sec:
            if self.is_clickable():
                return True
            time.sleep(0.01)
        return False

    def wait_to_be_clickable_for(self: ClickableMixinTypeT, timeout_sec: float) -> ClickableMixinTypeT:
        """
        This method waits for a maximum of `timeout_sec` seconds for the element to be clickable.

        :param timeout_sec: the maximum time in seconds to wait for this element to be clickable
        :raises TimeoutError: if timeout is exceeded
        """
        if not self.is_clickable_within(time_sec=timeout_sec):
            raise TimeoutError(f'element {self} is not clickable within {timeout_sec} seconds')
        return self



from abc import ABC, abstractmethod


class SelectByIndexMixin(ABC):
    """
    mixin class that describes that this element has selectable elements that can be selected-by-index
    """
    @abstractmethod
    def select_by_index(self, index: int) -> None:
        """
        This method selects the element by its index
        :param index: the index to select
        """



from __future__ import annotations
from typing import TypeVar
import enum
from abc import ABC


ConvertedBaseSelectorTypeT = TypeVar('ConvertedBaseSelectorTypeT', bound='BaseSelector')


class BaseSelector(ABC):
    """
    Base class that is used to identify GUI elements. It is normally overwritten in sub packages.
    """
    class By(enum.Enum):
        """
        Enum that describes the selecting type
        """

    class NoTranslationPossibleError(Exception):
        """
        raised when no translation is possible, because no translation was defined.
        """

    translations: dict[type[BaseSelector], dict[BaseSelector.By, BaseSelector.By]] = {}

    def __init__(self, by_type: By, identifier: str):
        self._by_type = by_type
        self._identifier = identifier

    def __eq__(self, other):
        return self.by_type == other.by_type and self.identifier == other.identifier

    def __str__(self):
        return f"Selector<{self._by_type}: {self._identifier}>"

    @property
    def by_type(self) -> By:
        """
        :return: the used selecting type
        """
        return self._by_type

    @property
    def identifier(self) -> str:
        """
        :return: the value of the selector, that is used to identify the element
        """
        return self._identifier

    def translate_to(self, other_selector_type: type[ConvertedBaseSelectorTypeT]) -> ConvertedBaseSelectorTypeT:
        """
        This method translates the given selector value to the corresponding selector value of the provided selector \
        type.

        :param other_selector_type: the selector type object this selector should translate to
        :return: the same selector value like this selector but as translated object from type given in
                 ``other_selector_type``
        """
        if not isinstance(other_selector_type, type) or not issubclass(other_selector_type, BaseSelector):
            raise TypeError(f'other_selector_type must be a subclass of `{BaseSelector.__name__}`')

        if self.__class__ == other_selector_type:
            return self

        translation_from_self = self.translations.get(other_selector_type)
        translation_from_other = other_selector_type.translations.get(self.__class__)
        reversed_translation_from_other = None
        if translation_from_other is not None:
            # reverse it
            reversed_translation_from_other = {v: k for k, v in translation_from_other.items()}
            if len(reversed_translation_from_other) != len(translation_from_other):
                raise KeyError('unable to reverse translation because some values are used more than once')

        if translation_from_self is None and translation_from_other is None:
            raise self.NoTranslationPossibleError(
                'can not translate selector because neither this selector has a translation nor the other element '
                'provides a translation for these specific types')

        result_from_self = translation_from_self[self.by_type] if translation_from_self is not None else None
        result_from_other = \
            reversed_translation_from_other[self.by_type] if reversed_translation_from_other is not None else None

        if result_from_self is not None and result_from_other is not None and result_from_self != result_from_other:
            raise self.NoTranslationPossibleError(
                f'both selector types define a own translation that return different results - {self.__class__} '
                f'returns `{result_from_self}` and {other_selector_type} returns `{result_from_other}`')

        new_by = result_from_self if translation_from_other is None else result_from_other
        return other_selector_type(new_by, self.identifier)



from abc import ABC, abstractmethod
from typing import TypeVar
import time

VisibleMixinTypeT = TypeVar("VisibleMixinTypeT", bound="VisibleMixin")


class VisibleMixin(ABC):
    """
    mixin class that describes a visible element
    """

    @abstractmethod
    def is_visible(self) -> bool:
        """
        This method returns True if the element is visible.

        .. note::
            This method also returns True if the element is theoretically visible but not visible because other
            elements cover it up.

        :return: True if the element is shown, otherwise it returns False
        """

    def is_visible_within(self, time_sec: float) -> bool:
        """
        This method waits for a maximum of `time_sec` seconds for the element to be visible. If the element is not
        shown within the provided time `time_sec`, the method returns False.

        :param time_sec: the maximum time to wait for this element to be visbile
        :return: True if it is visible, otherwise False
        """
        start_time = time.perf_counter()
        while time.perf_counter() - start_time < time_sec:
            if self.is_visible():
                return True
            time.sleep(0.01)
        return False

    def wait_to_be_visible_for(self: VisibleMixinTypeT, timeout_sec: float) -> VisibleMixinTypeT:
        """
        This method waits for a maximum of `timeout_sec` seconds for the element to be visible.

        :param timeout_sec: the maximum time in seconds to wait for this element to be visible
        :raises TimeoutError: if timeout is exceeded
        """
        if not self.is_visible_within(time_sec=timeout_sec):
            raise TimeoutError(f'element {self} was not visible within {timeout_sec} seconds')
        return self

    def was_hidden_within(self, time_sec: float) -> bool:
        """
        This method waits for a maximum of `time_sec` seconds for the hiddenness of this element. If the element is
        still visible after `time_sec`, the method returns False.

        :param time_sec: the maximum time to wait for this element to be hidden
        :return: True if it is not shown anymore (within the `time_sec`), otherwise False
        """
        start_time = time.perf_counter()
        while time.perf_counter() - start_time < time_sec:
            if not self.is_visible():
                return True
            time.sleep(0.01)
        return False

    def wait_to_be_hidden_for(self: VisibleMixinTypeT, timeout_sec: float) -> VisibleMixinTypeT:
        """
        This method waits for a maximum of `timeout_sec` seconds for the element to be hidden.

        :param timeout_sec: the maximum time to wait for this element to be hidden
        :raises TimeoutError: if timeout is exceeded
        """
        if not self.was_hidden_within(time_sec=timeout_sec):
            raise TimeoutError(f'element {self} was not hidden within {timeout_sec} seconds')
        return self



from abc import ABC, abstractmethod


class RectangleMixin(ABC):
    """
    mixin class that describes a rectangle object
    """

    @property
    @abstractmethod
    def width(self) -> float:
        """
        :return: returns the width of the element
        """

    @property
    @abstractmethod
    def height(self) -> float:
        """
        :return: returns the height of the element
        """



from abc import ABC, abstractmethod


class TwoStateCheckboxMixin(ABC):
    """
    mixin class that describes that this element is a two-state checkbox
    """

    @abstractmethod
    def is_checked(self) -> bool:
        """
        This method returns True if the element is checked, otherwise False
        :return: True if the element is checked, otherwise False
        """



from abc import ABC, abstractmethod


class TypeableMixin(ABC):
    """
    mixin class that describes that you can type text in this checkbox
    """
    @abstractmethod
    def type_text(self, text: str, clean_before: bool = False)-> None:
        """
        This method types a text in this element.

        :param text: the text that should be inserted
        :param clean_before: True if the system should make sure that all previous filled text is deleted before
                             inserting any new text
        """



from abc import ABC, abstractmethod


class HasValueMixin(ABC):
    """
    mixin class that describes an element that can have a value
    """

    @abstractmethod
    def get_value(self) -> object:
        """
        This method returns the value of the element.
        :return: the value of the element
        """



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TypeVar
import time

AbstractElementTypeT = TypeVar("AbstractElementTypeT", bound="AbstractElement")


class AbstractElement(ABC):
    """
    The base class for any kind of elements mostly used in another element or a :class:`PageFeature`
    """

    @abstractmethod
    def exists(self) -> bool:
        """
        Callback that checks if this element exists.

        .. note::
            The existence of an element does not necessarily mean that it visible.

        :return: True if it does exist, otherwise False
        """

    def exists_within(self, time_sec: float) -> bool:
        """
        This method waits for a maximum of `time_sec` seconds for the existence of this element. If the element does
        not exist within the provided time `time_sec`, the method returns False.

        :param time_sec: the maximum time to wait for this element to exist
        :return: True if it does exist, otherwise False
        """
        start_time = time.perf_counter()
        while time.perf_counter() - start_time < time_sec:
            if self.exists():
                return True
            time.sleep(0.01)
        return False

    def wait_to_exist_for(self: AbstractElementTypeT, timeout_sec: float) -> AbstractElementTypeT:
        """
        This method waits for a maximum of `timeout_sec` seconds for the existence of this element.

        :param timeout_sec: the maximum time in seconds to wait for this element to exist
        :raises TimeoutError: if timeout is exceeded
        """
        if not self.exists_within(time_sec=timeout_sec):
            raise TimeoutError(f'element {self} does not exist within {timeout_sec} seconds')
        return self

    def was_removed_within(self, time_sec: float) -> bool:
        """
        This method waits for a maximum of `time_sec` seconds for the removement of this element. If the element still
        exists after `time_sec`, the method returns False.

        :param time_sec: the maximum time to wait for this element to be removed
        :return: True if it does not exist anymore (within the `time_sec`), otherwise False
        """
        start_time = time.perf_counter()
        while time.perf_counter() - start_time < time_sec:
            if not self.exists():
                return True
            time.sleep(0.01)
        return False

    def wait_to_be_removed_for(self: AbstractElementTypeT, timeout_sec: float) -> AbstractElementTypeT:
        """
        This method waits for a maximum of `timeout_sec` seconds for the removement of this element.

        :param timeout_sec: the maximum time to wait for this element to be removed
        :raises TimeoutError: if timeout is exceeded
        """
        if not self.was_removed_within(time_sec=timeout_sec):
            raise TimeoutError(f'element {self} was not removed within {timeout_sec} seconds')
        return self



from abc import ABC, abstractmethod


class ScrollableMixin(ABC):
    """
    mixin class that describes that the content of this element can be scrolled
    """
    @abstractmethod
    def scroll_to_beginning(self) -> None:
        """
        This method scrolls to the beginning of the scrollable area
        """

    @abstractmethod
    def scroll_once_forward(self) -> None:
        """
        This method scrolls for one element forward
        """

    @abstractmethod
    def scroll_once_backward(self) -> None:
        """
        This method scrolls for one element backward
        """

    @abstractmethod
    def scroll_to_end(self) -> None:
        """
        This method scrolls to the end of the scrollable area
        """



import time
from balder import Feature


class PageFeature(Feature):
    """
    Base class that represents a page. This can be any type of page.
    """

    def is_applicable(self) -> bool:
        """
        Checks if this specific page can be used on the current screen.
        :return: True if this page object can be applied on the current screen, otherwise False
        """
        raise NotImplementedError

    def wait_for_page(self, timout_sec=10) -> None:
        """
        This method waits for a specific time to make sure that this page can be applied on the current screen.

        :param timout_sec: the maximum time in seconds to wait
        :raise TimeoutError: if the page cannot be applied
        """
        start_time = time.time()
        while time.time() - start_time < timout_sec:
            if self.is_applicable():
                return
        raise TimeoutError('timeout waiting for page to be visible')



from abc import ABC, abstractmethod
from .list_container_mixin import ListContainerMixin

class PaginatedContainerMixin(ListContainerMixin, ABC):
    """
    This mixin is used for list container with a pagination functionality
    """

    class PageDoesNotExist(Exception):
        """
        This exception is raised when a page does not exist.
        """

    @property
    @abstractmethod
    def current_page(self) -> int:
        """
        :return: returns the current page number
        """

    @property
    @abstractmethod
    def min_max_page(self) -> tuple[int, int]:
        """
        :return: returns the min and the max page that exists for this paginated list
        """

    @abstractmethod
    def go_to_page(self, page: int) -> None:
        """
        This method goes to the provided page.
        :param page: the page index to go
        :raises PageDoesNotExist: if the page provided page does not exist
        """

    @abstractmethod
    def go_to_next_page(self):
        """
        This method goes to the next page
        :raises PageDoesNotExist: if the current page is the latest already
        """

    @abstractmethod
    def go_to_previous_page(self):
        """
        This method goes to the previous page
        :raises PageDoesNotExist: if the current page is the latest already
        """



from abc import ABC, abstractmethod


class CircleMixin(ABC):
    """
    mixin class that describes a circle object
    """

    @property
    @abstractmethod
    def diameter(self) -> float:
        """
        :return: returns the circle diameter
        """



from .circle_mixin import CircleMixin
from .clickable_mixin import ClickableMixin
from .has_value_mixin import HasValueMixin
from .list_container_mixin import ListContainerMixin
from .paginated_container_mixin import PaginatedContainerMixin
from .rectangle_mixin import RectangleMixin
from .scrollable_mixin import ScrollableMixin
from .select_by_hidden_value_mixin import SelectByHiddenValueMixin
from .select_by_index_mixin import SelectByIndexMixin
from .select_by_visible_text_mixin import SelectByVisibleTextMixin
from .two_state_checkbox_mixin import TwoStateCheckboxMixin
from .typeable_mixin import TypeableMixin
from .visible_mixin import VisibleMixin

__all__ = [
    "CircleMixin",
    "ClickableMixin",
    "HasValueMixin",
    "ListContainerMixin",
    "PaginatedContainerMixin",
    "RectangleMixin",
    "ScrollableMixin",
    "SelectByHiddenValueMixin",
    "SelectByIndexMixin",
    "SelectByVisibleTextMixin",
    "TwoStateCheckboxMixin",
    "TypeableMixin",
    "VisibleMixin",
]



from typing import List
from abc import ABC, abstractmethod
from ..components.abstract_element import AbstractElement


class ListContainerMixin(ABC):
    """
    mixin class that describes that the element is a list-container
    """

    @abstractmethod
    def get_child_elements(self) -> List[AbstractElement]:
        """
        This method returns a list of elements that are contained within this container.
        :return: the list of elements that are contained within this container.
        """
